/* GSTEngineHandler.cpp */

/* Copyright (C) 2014  Lucio Carreras
 *
 * This file is part of sayonara player
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY{} without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "EngineHandler.h"
#include "Playback/PlaybackEngine.h"
#include "Convert/ConvertEngine.h"


GSTEngineHandler::GSTEngineHandler(QObject* parent) : Engine(parent) {
	_cur_engine = nullptr;
	_play_manager = PlayManager::getInstance();

	GSTPlaybackEngine* pb_engine = new GSTPlaybackEngine();
	GSTConvertEngine* cvt_engine = new GSTConvertEngine();

	if(pb_engine->init()){
		_engines.push_back(pb_engine);
	}

	if(cvt_engine->init()){
		_engines.push_back(cvt_engine);
	}

	connect(_play_manager, &PlayManager::sig_playstate_changed,
			this, &GSTEngineHandler::playstate_changed);

	connect(_play_manager, &PlayManager::sig_track_changed,
			this, GSTEngineHandler_change_track_md);

	connect(_play_manager, &PlayManager::sig_seeked_abs_ms,
			this, &GSTEngineHandler::jump_abs_ms);

	connect(_play_manager, &PlayManager::sig_seeked_rel,
			this, &GSTEngineHandler::jump_rel);

	connect(_play_manager, &PlayManager::sig_record,
			this, &GSTEngineHandler::record_button_toggled);

	MetaData md = _play_manager->get_cur_track();
	if(!md.filepath().isEmpty()){
		change_track(md);
	}

	psl_change_engine(PLAYBACK_ENGINE);
}


GSTEngineHandler::~GSTEngineHandler() {
	
}

bool GSTEngineHandler::init() {

	return true;
}


void GSTEngineHandler::start_convert() {

	stop();

	if( QString::compare(_cur_engine->get_name(), CONVERT_ENGINE ) != 0) {
		psl_change_engine(CONVERT_ENGINE);
    }
}

void GSTEngineHandler::end_convert() {

	stop();

	sp_log(Log::Debug) << "Engine end convert";
	if( QString::compare(_cur_engine->get_name(), PLAYBACK_ENGINE ) != 0) {
		sp_log(Log::Debug) << "Change to playback engine";
		psl_change_engine(PLAYBACK_ENGINE);
	}
}

void GSTEngineHandler::psl_set_speed(float f) {
	if(!_cur_engine) return;
	_cur_engine->psl_set_speed(f);
}

void GSTEngineHandler::fill_engines(const QVector<Engine*>& engines) {
	_engines = engines;

	psl_change_engine(PLAYBACK_ENGINE);
}


void GSTEngineHandler::playstate_changed(PlayManager::PlayState state){

	if(!_cur_engine) return;

	switch(state){
		case PlayManager::PlayState::Playing:
			play();
			break;

		case PlayManager::PlayState::Paused:
			pause();
			break;

		case PlayManager::PlayState::Stopped:
			stop();
			break;

		default:
			return;
	}
}


void GSTEngineHandler::play() {
	if(!_cur_engine) return;

	_cur_engine->play();
}

void GSTEngineHandler::stop() {

	if(!_cur_engine) return;
	_cur_engine->stop();
}

void GSTEngineHandler::pause() {
	if(!_cur_engine) return;
	_cur_engine->pause();
}



void GSTEngineHandler::jump_abs_ms(quint64 ms) {
	if(!_cur_engine) return;
	_cur_engine->jump_abs_ms(ms);
}

void GSTEngineHandler::jump_rel_ms(quint64 ms){
	if(!_cur_engine) return;
	_cur_engine->jump_rel_ms(ms);
}

void GSTEngineHandler::jump_rel(double where) {
	if(!_cur_engine) return;
	_cur_engine->jump_rel(where);
}


void GSTEngineHandler::change_track(const MetaData& md) {
	if(!_cur_engine) return;
	_cur_engine->change_track(md);
}

void GSTEngineHandler::change_track(const QString& str) {
	if(!_cur_engine) return;
	_cur_engine->change_track(str);
}

void GSTEngineHandler::eq_changed(int band, int value) {
	if(!_cur_engine) return;
	_cur_engine->eq_changed(band, value);
}


void GSTEngineHandler::record_button_toggled(bool b) {
	if(!_cur_engine) return;
	_cur_engine->record_button_toggled(b);
}


void GSTEngineHandler::psl_new_stream_session() {
	if(!_cur_engine) return;
	_cur_engine->psl_new_stream_session();
}

void GSTEngineHandler::sl_md_changed(const MetaData& v) {
	emit sig_md_changed(v);
}

void GSTEngineHandler::sl_dur_changed(const MetaData& md){
	emit sig_dur_changed(md);
}

void GSTEngineHandler::sl_pos_changed_ms(quint64 v) {
	_play_manager->set_position_ms( v);
}

void GSTEngineHandler::sl_pos_changed_s(quint32 v) {
	_play_manager->set_position_ms( (quint64) (v * 1000) );
}


void GSTEngineHandler::sl_track_finished() {
	_play_manager->next();
}

void GSTEngineHandler::set_track_ready(){
	_play_manager->set_track_ready();
}

void GSTEngineHandler::buffering(int progress){
	_play_manager->buffer(progress);
}

void GSTEngineHandler::sl_level(float l, float r) {
	emit sig_level(l, r);
}

void GSTEngineHandler::sl_spectrum(const QList<float>& lst) {
	emit sig_spectrum(lst);
}


bool GSTEngineHandler::configure_connections(Engine* old_engine, Engine* new_engine) {

	if(!old_engine && !new_engine) return false;
	if(old_engine == new_engine) return false;

	if(old_engine) {
		disconnect(old_engine, &Engine::sig_track_ready, this, &GSTEngineHandler::set_track_ready);
		disconnect(old_engine, &Engine::sig_md_changed, this, &GSTEngineHandler::sl_md_changed);
		disconnect(old_engine, &Engine::sig_pos_changed_ms, this, &GSTEngineHandler::sl_pos_changed_ms);
		disconnect(old_engine, &Engine::sig_pos_changed_s, this, &GSTEngineHandler::sl_pos_changed_s);
		disconnect(old_engine, &Engine::sig_dur_changed, this, &GSTEngineHandler::sl_dur_changed);
		disconnect(old_engine, &Engine::sig_br_changed, this, &Engine::sig_br_changed);
		disconnect(old_engine, &Engine::sig_track_finished, this, &GSTEngineHandler::sl_track_finished);
		disconnect(old_engine, &Engine::sig_level, this, &GSTEngineHandler::sl_level);
		disconnect(old_engine, &Engine::sig_spectrum, this, &GSTEngineHandler::sl_spectrum);
		disconnect(old_engine, &Engine::sig_data, this, &GSTEngineHandler::new_data);
		disconnect(old_engine, &Engine::sig_buffering, this, &GSTEngineHandler::buffering);
	}

	if(new_engine) {
		connect(new_engine, &Engine::sig_track_ready, this, &GSTEngineHandler::set_track_ready);
		connect(new_engine, &Engine::sig_md_changed, this, &GSTEngineHandler::sl_md_changed);
		connect(new_engine, &Engine::sig_pos_changed_ms, this, &GSTEngineHandler::sl_pos_changed_ms);
		connect(new_engine, &Engine::sig_pos_changed_s, this, &GSTEngineHandler::sl_pos_changed_s);
		connect(new_engine, &Engine::sig_dur_changed, this, &GSTEngineHandler::sl_dur_changed);
		connect(new_engine, &Engine::sig_br_changed, this, &Engine::sig_br_changed);
		connect(new_engine, &Engine::sig_track_finished, this, &GSTEngineHandler::sl_track_finished);
		connect(new_engine, &Engine::sig_level, this, &GSTEngineHandler::sl_level);
		connect(new_engine, &Engine::sig_spectrum, this, &GSTEngineHandler::sl_spectrum);
		connect(new_engine, &Engine::sig_data, this, &GSTEngineHandler::new_data);
		connect(new_engine, &Engine::sig_buffering, this, &GSTEngineHandler::buffering);
	}

	return true;
}


void GSTEngineHandler::psl_change_engine(QString name) {
	Engine* new_engine=nullptr;

	for(Engine* engine : _engines) {
		if(engine && name.compare(engine->get_name(), Qt::CaseInsensitive) == 0) {
			new_engine = engine;
			break;
		}		
	}

	configure_connections(_cur_engine, new_engine);
	_cur_engine = new_engine;

}

