
/* Copyright (C) 2011-2015  Lucio Carreras
 *
 * This file is part of sayonara player
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "ModifierCallbacks.h"
#include "ElementOperator.h"


static bool send_sink_eos(GstElement* element){
	bool success = false;
	GstPad* sink_pad = gst_element_get_static_pad( element, "sink" );

	if(sink_pad){
		success = gst_pad_send_event( sink_pad, gst_event_new_eos() );
		gst_object_unref(sink_pad);
	}

	return success;
}

static bool wait_for_eos(GstElement* element, gpointer userdata, GstPadProbeCallback cb){

	GstPad* src_pad = gst_element_get_static_pad( element, "src" );

	if(src_pad){

		gst_pad_add_probe(src_pad,
						  (GstPadProbeType)(GST_PAD_PROBE_TYPE_BLOCK | GST_PAD_PROBE_TYPE_EVENT_DOWNSTREAM),
						  cb,
						  userdata,
						  NULL
		);

		gst_object_unref(src_pad);
	}

	return (src_pad != nullptr);
}


GstPadProbeReturn
ModifierCallbacks::remove_eos_received(GstPad *pad, GstPadProbeInfo *info, gpointer user_data){

	ElementOperator* op = static_cast<ElementOperator*>(user_data);
	op->set_status(Q_FUNC_INFO);

	gst_pad_remove_probe( pad, GST_PAD_PROBE_INFO_ID (info) );

	PipelineHelper::remove_elements(op->pipeline(), op->elems_of_interest());

	ElementVector vec;
	vec << op->predecessor()
		<< op->successor();

	PipelineHelper::link_elements(vec);

	op->set_status(QString(Q_FUNC_INFO) + " ended");
	return GST_PAD_PROBE_DROP;
}


GstPadProbeReturn
ModifierCallbacks::remove_predecessor_blocked(GstPad* pad, GstPadProbeInfo* info, gpointer user_data){

	ElementOperator* op = static_cast<ElementOperator*> (user_data);
	op->set_status(Q_FUNC_INFO);

	gst_pad_remove_probe(pad, GST_PAD_PROBE_INFO_ID(info));

	wait_for_eos(op->elem_of_interest(op->elems_of_interest().size() - 1), user_data, remove_eos_received);
	send_sink_eos(op->elem_of_interest(0));

	op->set_status(QString(Q_FUNC_INFO) + " ended");
	return GST_PAD_PROBE_OK;
}


GstPadProbeReturn
ModifierCallbacks::replace_eos_received(GstPad* pad, GstPadProbeInfo* info, gpointer user_data){

	ElementOperator* op = static_cast<ElementOperator*>(user_data);
	op->set_status(Q_FUNC_INFO);

	if(op->has_probe()){
		gst_pad_remove_probe(pad, GST_PAD_PROBE_INFO_ID(info));
	}

	ElementVector to_link;
	to_link << op->predecessor()
			<< op->elem_of_interest(1)
			<< op->successor();


	PipelineHelper::remove_element(op->pipeline(), op->elem_of_interest(0));
	PipelineHelper::add_element(op->pipeline(), op->elem_of_interest(1));
	PipelineHelper::link_elements(to_link);
	PipelineHelper::apply_state(op->pipeline(), op->elem_of_interest(1));

	op->set_status(QString(Q_FUNC_INFO) + " ended");
	return GST_PAD_PROBE_DROP;
}


GstPadProbeReturn
ModifierCallbacks::replace_predecessor_blocked(GstPad* pad, GstPadProbeInfo* info, gpointer user_data){

	gst_pad_remove_probe(pad, GST_PAD_PROBE_INFO_ID(info));

	ElementOperator* op = static_cast<ElementOperator*> (user_data);
	op->set_status(Q_FUNC_INFO);
	bool success = wait_for_eos(op->elem_of_interest(0), user_data, replace_eos_received);

	// this element is probably the last one in the pipeline
	if(!success){
		op->set_status(QString(Q_FUNC_INFO) + " finished 1");
		op->set_probe(false);
		return replace_eos_received(pad, info, user_data);
	}

	op->set_probe(true);
	send_sink_eos(op->elem_of_interest(0));

	op->set_status(QString(Q_FUNC_INFO) + " ended");
	return GST_PAD_PROBE_OK;
}



GstPadProbeReturn
ModifierCallbacks::add_predecessor_blocked(GstPad *pad, GstPadProbeInfo *info, gpointer user_data){

	bool success;
	ElementOperator* op = static_cast<ElementOperator*>(user_data);
	op->set_status(Q_FUNC_INFO);

	gst_pad_remove_probe( pad, GST_PAD_PROBE_INFO_ID (info) );
	gst_element_unlink(op->predecessor(), op->successor() );

	gst_element_set_state(op->elem_of_interest(0), GST_STATE_NULL);
	success = PipelineHelper::add_elements(op->pipeline(), op->elems_of_interest());
	if(!success){
		op->set_status(QString(Q_FUNC_INFO) + " finished 1");
		return GST_PAD_PROBE_OK;
	}

	ElementVector vec;
	vec << op->predecessor()
		<< op->elems_of_interest()
		<< op->successor();

	PipelineHelper::link_elements(vec);
	PipelineHelper::apply_state(op->pipeline(), op->elems_of_interest());

	op->set_status(QString(Q_FUNC_INFO) + " ended");
	return GST_PAD_PROBE_OK;
}


