/* PipelineModifier.cpp */

/* Copyright (C) 2011-2015  Lucio Carreras
 *
 * This file is part of sayonara player
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "PipelineModifier.h"
#include "ElementOperator.h"
#include <QMetaType>


PipelineModifier::PipelineModifier(GstElement* pipeline){

	qRegisterMetaType<PipelineModifier::Status>("PipelineModifier::Status");

	_pipeline = pipeline;
	_status = Status::Finished;
	_operator = nullptr;
}

PipelineModifier::~PipelineModifier(){

}


bool PipelineModifier::add(GstElement *predecessor, const ElementVector& elems_of_interest, GstElement *successor){

	if(_status == Status::Adding){
		sp_log(Log::Debug) << "Already Adding";
		return false;
	}

	else if(_status != Status::Finished){
		sp_log(Log::Debug) << "Sorry, but current status = " << (int) _status;
		return false;
	}

	_status = Status::Adding;

	gchar* name = gst_object_get_name(GST_OBJECT(elems_of_interest[0]));
	GstElement* p = gst_bin_get_by_name(GST_BIN(_pipeline), name);

	if( p != nullptr){
		_status = Status::Finished;
		set_finished();
		return true;
	}

	_operator = new ElementOperator(this, _pipeline, ElementOperator::Mode::Add);
	_operator->set_predecessor(predecessor);
	_operator->set_successor(successor);
	_operator->set_elems_of_interest(elems_of_interest);
	_operator->start();

	return true;
}


bool PipelineModifier::remove(GstElement* predecessor, const QVector<GstElement*>& elems_of_interest, GstElement* successor){

	if(_status == Status::Removing){
		sp_log(Log::Debug) << "Already removing";
		return false;
	}

	else if(_status != Status::Finished){
		sp_log(Log::Debug) << "Sorry, but current status = " << (int) _status;
		return false;
	}

	_status = Status::Removing;
	gchar* name = GST_ELEMENT_NAME( elems_of_interest[0] );

	if( gst_bin_get_by_name(GST_BIN(_pipeline), name) == nullptr){
		_status = Status::Finished;
		set_finished();
		return true;
	}

	_operator = new ElementOperator(this, _pipeline, ElementOperator::Mode::Remove);
	_operator->set_predecessor(predecessor);
	_operator->set_successor(successor);
	_operator->set_elems_of_interest(elems_of_interest);
	_operator->start();

	return true;
}


bool PipelineModifier::replace(GstElement* predecessor, GstElement *org_element, GstElement *new_element, GstElement* successor){


	if(_status == Status::Replacing){
		sp_log(Log::Debug) << "Already Replacing";
		return false;
	}

	else if(_status != Status::Finished){
		sp_log(Log::Debug) << "Sorry, but current status = " << (int) _status;
		return false;
	}

	_status = Status::Replacing;
	gchar* name = GST_ELEMENT_NAME( org_element );

	if( gst_bin_get_by_name(GST_BIN(_pipeline), name) == nullptr){
		_status = Status::Finished;
		set_finished();
		return true;
	}

	_operator = new ElementOperator(this, _pipeline, ElementOperator::Mode::Replace);
	_operator->set_predecessor(predecessor);
	_operator->set_successor(successor);
	_operator->set_elems_of_interest(ElementVector() << org_element << new_element);
	_operator->start();

	return true;
}

void PipelineModifier::set_finished(){

	PipelineModifier::Status status = _status;
	_status = Status::Finished;

	emit sig_finished(status);

	if(_operator != nullptr){
		delete _operator;
	}

	_operator = nullptr;
}


PipelineModifier::Status PipelineModifier::get_status(){
	return _status;
}
