/* GSTPlaybackPipeline.cpp */

/* Copyright (C) 2011-2015  Lucio Carreras
 *
 * This file is part of sayonara player
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "PlaybackPipeline.h"




GSTPlaybackPipeline::GSTPlaybackPipeline(Engine* engine, QObject *parent) :
	GSTAbstractPipeline("Playback Pipeline", engine, parent)
{
	_speed_val = 1.0f;
	_speed_active = false;

	_speed_active = false;
	_speed_val = 1.0f;
	_vol = 0;

	_timer = new QTimer();
	_timer->setInterval(5);

	_pm_lame = nullptr;
	_pm_level = nullptr;
	_pm_spectrum = nullptr;

	connect(_timer, &QTimer::timeout, this, &GSTPlaybackPipeline::play);
}


GSTPlaybackPipeline::~GSTPlaybackPipeline() {

	delete _timer;
}

bool GSTPlaybackPipeline::init(GstState state){

	if(!GSTAbstractPipeline::init(state)){
		return false;
	}

	_pm_lame = new PipelineModifier(_pipeline);
	connect(_pm_lame, &PipelineModifier::sig_finished,
			this, &GSTPlaybackPipeline::lame_pipeline_modified);


	_pm_level = new PipelineModifier(_pipeline);
	_pm_spectrum = new PipelineModifier(_pipeline);

	_settings->set(SetNoDB::MP3enc_found, _lame != nullptr);

	REGISTER_LISTENER(Set::Engine_Vol, _sl_vol_changed);
	REGISTER_LISTENER(SetNoDB::Broadcast_Clients, _sl_broadcast_clients_changed);

	// set by gui, initialized directly in pipeline
	REGISTER_LISTENER_NO_CALL(Set::Engine_ShowLevel, _sl_show_level_changed);
	REGISTER_LISTENER_NO_CALL(Set::Engine_ShowSpectrum, _sl_show_spectrum_changed);


	return true;
}

bool GSTPlaybackPipeline::create_elements(){

	// input
	if(!create_element(&_audio_src, "uridecodebin", "src")) return false;
	if(!create_element(&_audio_convert, "audioconvert")) return false;
	if(!create_element(&_equalizer, "equalizer-10bands")) return false;
	if(!create_element(&_tee, "tee")) return false;

	// standard output branch
	if(!create_element(&_eq_queue, "queue", "eq_queue")) return false;
	if(!create_element(&_volume, "volume")) return false;
	if(!create_element(&_audio_sink, "autoaudiosink")) return false;

	// level branch
	if(!create_element(&_level_queue, "queue", "level_queue")) return false;
	if(!create_element(&_level, "level")) return false;
	if(!create_element(&_level_sink, "fakesink", "level_sink")) return false;

	// spectrum branch
	if(!create_element(&_spectrum_queue, "queue", "spectrum_queue")) return false;
	if(!create_element(&_spectrum, "spectrum")) return false;
	if(!create_element(&_spectrum_sink,"fakesink", "spectrum_sink")) return false;

	// lame branch
	if( !create_element(&_lame_queue, "queue", "lame_queue") ||
		!create_element(&_lame_converter, "audioconvert", "lame_converter") ||
		!create_element(&_lame_resampler, "audioresample", "lame_resampler") ||
		!create_element(&_lame, "lamemp3enc") ||
		!create_element(&_lame_app_sink, "appsink", "lame_appsink") ||
		!create_element(&_lame_fake_sink, "fakesink", "lame_fakesink"))
	{
		_lame = nullptr;
	}

	// stream recorder branch
	if(	!create_element(&_file_queue, "queue", "sr_queue") ||
		!create_element(&_file_resampler, "audioresample", "sr_resample") ||
		!create_element(&_file_lame, "lamemp3enc", "sr_lame")  ||
		!create_element(&_file_sink, "filesink", "sr_filesink") ||
		!create_element(&_file_fake_sink, "fakesink","sr_fakesink"))
	{
		_file_sink = nullptr;
	}

	return true;
}

bool GSTPlaybackPipeline::add_and_link_elements(){

	bool success = false;
	GstPadTemplate* tee_src_pad_template;

	gst_bin_add_many(GST_BIN(_pipeline),
					 _audio_src, _audio_convert, _equalizer, _tee,

					 _eq_queue, _volume, _audio_sink,
					 _level_queue, _level_sink,
					 _spectrum_queue, _spectrum_sink,

					 nullptr);

	/* before tee */
	success = gst_element_link_many(_audio_convert, _equalizer, _tee, nullptr);
	if(!_test_and_error_bool(success, "Engine: Cannot link audio convert with tee")){
		return false;
	}


	/* standard output branch */
	success = gst_element_link_many(_eq_queue, _volume, _audio_sink, nullptr);
	if(!_test_and_error_bool(success, "Engine: Cannot link eq with audio sink")) {
		return false;
	}


	/* level branch */
	if(_settings->get(Set::Engine_ShowLevel)){
		gst_bin_add(GST_BIN(_pipeline), _level);
		success = gst_element_link_many(_level_queue, _level, _level_sink, nullptr);
	}
	else{
		success = gst_element_link_many(_level_queue, _level_sink, nullptr);
	}

	if(!_test_and_error_bool(success, "Engine: Cannot link Level pipeline")){
		return false;
	}


	/* spectrum branch */
	if(_settings->get(Set::Engine_ShowSpectrum)){
		gst_bin_add(GST_BIN(_pipeline), _spectrum);
		success = gst_element_link_many(_spectrum_queue, _spectrum, _spectrum_sink, nullptr);
	}
	else {
		success = gst_element_link_many(_spectrum_queue, _spectrum_sink, nullptr);
	}

	if(!_test_and_error_bool(success, "Engine: Cannot link Spectrum pipeline")){
		return false;
	}


	/* lame branch (optional) */
	if(_lame){
		gst_bin_add_many(GST_BIN(_pipeline), _lame_queue, _lame_fake_sink, nullptr);
		success = gst_element_link_many( _lame_queue, _lame_fake_sink, nullptr);
		_test_and_error_bool(success, "Engine: Cannot link lame stuff");
	}

	/* stream rippper branch (optional) */
	if(_file_sink){
		gst_bin_add_many(GST_BIN(_pipeline), _file_queue, _file_fake_sink, nullptr);
		success = gst_element_link_many( _file_queue, _file_fake_sink, nullptr);
		_test_and_error_bool(success, "Engine: Cannot link streamripper stuff");
	}

	/* create tee pads */
	tee_src_pad_template = gst_element_class_get_pad_template (GST_ELEMENT_GET_CLASS (_tee), "src_%u");
	if(!_test_and_error(tee_src_pad_template, "Engine: _tee_src_pad_template is nullptr")) {
		return false;
	}

	/* connect branches with tee */
	success = tee_connect(tee_src_pad_template, _level_queue, "Level");
	if(!_test_and_error_bool(success, "Engine: Cannot link level queue with tee")){
		return false;
	}

	tee_connect(tee_src_pad_template, _spectrum_queue, "Spectrum");
	if(!_test_and_error_bool(success, "Engine: Cannot link spectrum queue with tee")){
		return false;
	}

	tee_connect(tee_src_pad_template, _eq_queue, "Equalizer");
	if(!_test_and_error_bool(success, "Engine: Cannot link eq queue with tee")){
		return false;
	}

	if(_lame){
		success = tee_connect(tee_src_pad_template, _lame_queue, "Lame");
		if(!_test_and_error_bool(success, "Engine: Cannot link lame queue with tee")){
			_settings->set(SetNoDB::MP3enc_found, false);
		}
	}

	if(_file_sink){
		success = tee_connect(tee_src_pad_template, _file_queue, "Streamripper");
		if(!_test_and_error_bool(success, "Engine: Cannot link streamripper stuff")){
			_settings->set(Set::Engine_SR_Active, false);
		}
	}

	return true;
}

bool GSTPlaybackPipeline::configure_elements(){

	guint64 interval = 25000000;
	gint threshold = -crop_spectrum_at;
	QList<GstElement*> sinks;

	g_object_set (G_OBJECT (_audio_src),
				  "use-buffering", true,
				  "buffer-size", 100000, // 100kB
				  nullptr);


	g_object_set (G_OBJECT (_level),
				  "message", true,
				  "interval", interval,
				  nullptr);

	g_object_set (G_OBJECT (_spectrum),
				  "interval", interval,
				  "bands", N_BINS,
				  "threshold", threshold,
				  "message-phase", false,
				  "message-magnitude", true,
				  "multi-channel", false,
				  nullptr);

	/* run synced and not as fast as we can */
	g_object_set(G_OBJECT (_audio_sink),
				 "sync",
				 true,
				 nullptr);

	if(_lame){
		g_object_set(G_OBJECT (_lame),
					 "perfect-timestamp", true,
					 "target", 1,
					 "cbr", true,
					 "bitrate", 128,
					 "encoding-engine-quality", 2,
					 nullptr);

		g_object_set(G_OBJECT(_lame_app_sink),
					 "emit-signals",
					 true, nullptr );

		gst_object_ref(_lame);
		gst_object_ref(_lame_resampler);

		sinks << _lame_app_sink << _lame_fake_sink;
	}

	if(_file_sink){
		g_object_set(G_OBJECT (_file_lame),

					 "perfect-timestamp", true,
					 "target", 1,
					 "cbr", true,
					 "bitrate", 128,
					 "encoding-engine-quality", 2,
					 nullptr);

		sinks << _file_sink << _file_fake_sink;
	}

	sinks << _level_sink << _spectrum_sink;

	for(GstElement* sink : sinks){

		/* run synced and not as fast as we can */
		g_object_set(G_OBJECT (sink), "sync", true, nullptr);

		/* change the state, when it's said to change the state */
		g_object_set(G_OBJECT (sink), "async", false, nullptr);
	}

	g_signal_connect (_audio_src, "pad-added", G_CALLBACK (PipelineCallbacks::pad_added_handler), _audio_convert);
	if(_lame){
		g_signal_connect (_lame_app_sink, "new-sample", G_CALLBACK(PipelineCallbacks::new_buffer), this);
	}

	return true;
}

bool GSTPlaybackPipeline::tee_connect(GstPadTemplate* tee_src_pad_template, GstElement* queue, const QString& queue_name){

	GstPadLinkReturn s;
	GstPad* tee_queue_pad;
	GstPad* queue_pad;

	QString error_1 = QString("Engine: Tee-") + queue_name + " pad is nullptr";
	QString error_2 = QString("Engine: ") + queue_name + " pad is nullptr";
	QString error_3 = QString("Engine: Cannot link tee with ") + queue_name;

	tee_queue_pad = gst_element_request_pad(_tee, tee_src_pad_template, nullptr, nullptr);
	if(!_test_and_error(tee_queue_pad, error_1)){
		return false;
	}

	queue_pad = gst_element_get_static_pad(queue, "sink");
	if(!_test_and_error(queue_pad, error_2)) {
		return false;
	}

	s = gst_pad_link (tee_queue_pad, queue_pad);
	if(!_test_and_error_bool((s == GST_PAD_LINK_OK), error_3)) {
		return false;
	}

	g_object_set (queue, "silent", TRUE, nullptr);

	return true;
}


void GSTPlaybackPipeline::play() {

	g_timeout_add(200, (GSourceFunc) PipelineCallbacks::position_changed, this);
	gst_element_set_state(_pipeline, GST_STATE_PLAYING);

	_timer->stop();

	if(_speed_active){
		set_speed(_speed_val);
	}
}


void GSTPlaybackPipeline::pause() {

	gst_element_set_state(_pipeline, GST_STATE_PAUSED);
}


void GSTPlaybackPipeline::stop() {

	_timer->stop();

	_duration_ms = 0;
	_uri = nullptr;

	gst_element_set_state(_pipeline	, GST_STATE_NULL);
}

void GSTPlaybackPipeline::_sl_vol_changed() {

	_vol = _settings->get(Set::Engine_Vol);
	if(_vol < 0) _vol = 0;
	if(_vol > 100) _vol = 100;

	float vol_val = (float) (_vol * 1.0f / 100.0f);

	g_object_set(G_OBJECT(_volume), "volume", vol_val, nullptr);
}


void GSTPlaybackPipeline::unmute() {

	g_object_set(G_OBJECT(_volume), "mute", FALSE, nullptr);
}


bool GSTPlaybackPipeline::_seek(gint64 ns){

	bool success;

	if(_speed_active)
	{
		success = gst_element_seek(_audio_src,
								   _speed_val,
								   GST_FORMAT_TIME,
								   (GstSeekFlags)(GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_SKIP),
								   GST_SEEK_TYPE_SET, ns,
								   GST_SEEK_TYPE_SET, _duration_ms * GST_MSECOND);

	}

	else
	{
		success = gst_element_seek_simple( _audio_src,
										   GST_FORMAT_TIME,
										   (GstSeekFlags) (GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_SNAP_NEAREST),
										   ns);
	}

	return success;
}

gint64 GSTPlaybackPipeline::seek_rel(double percent, gint64 ref_ns) {

	gint64 new_time_ns;

	if (percent > 1.0){
		new_time_ns = ref_ns;
	}

	else if(percent < 0){
		new_time_ns = 0;
	}

	else{
		new_time_ns = (percent * ref_ns); // nsecs
	}


	if( _seek(new_time_ns) ) {
		return new_time_ns;
	}

	return 0;
}


gint64 GSTPlaybackPipeline::seek_abs(gint64 ns) {

	if(ns < 0) ns = 0;

	if( _seek(ns) ) {
		return ns;
	}

	return 0;
}

void GSTPlaybackPipeline::set_speed(float f) {

	if(f < 0 && _speed_active) {

		_speed_active = false;
		_speed_val = 1.0f;
	}

	else if(f > 0 && !_speed_active) {

		_speed_active = true;
		_speed_val = f;

		_seek(_position_ms * GST_MSECOND);
	}

	else if(f > 0 && _speed_active ) {

		_speed_val = f;

		_seek(_position_ms * GST_MSECOND);
	}
}


void GSTPlaybackPipeline::_sl_show_level_changed() {

	bool active = _settings->get(Set::Engine_ShowLevel);

	if(active){
		_pm_level->add(_level_queue, ElementVector() << _level, _level_sink);
	}
	else{
		_pm_level->remove(_level_queue, ElementVector() << _level, _level_sink);
	}
}


void GSTPlaybackPipeline::_sl_show_spectrum_changed() {

	bool active = _settings->get(Set::Engine_ShowSpectrum);

	if(active){
		_pm_spectrum->add(_spectrum_queue, ElementVector() << _spectrum, _spectrum_sink);
	}
	else{
		_pm_spectrum->remove(_spectrum_queue, ElementVector() << _spectrum, _spectrum_sink);
	}
}


void GSTPlaybackPipeline::_sl_broadcast_clients_changed(){

	if(!_lame){
		return;
	}

	int n_clients = _settings->get(SetNoDB::Broadcast_Clients);
	bool active = (n_clients > 0);

	ElementVector elems_of_interest;
	elems_of_interest << _lame_converter << _lame_resampler << _lame;

	if(active){
		_pm_lame->add(_lame_queue, elems_of_interest, _lame_fake_sink);
	}

	else{
		_pm_lame->remove(_lame_queue, elems_of_interest, _lame_app_sink);
	}
}


void GSTPlaybackPipeline::lame_pipeline_modified(PipelineModifier::Status old_status){
	if(old_status == PipelineModifier::Status::Adding){
		_pm_lame->replace(_lame, _lame_fake_sink, _lame_app_sink, nullptr);
	}

	else if(old_status == PipelineModifier::Status::Removing){
		_pm_lame->replace(_lame_queue, _lame_app_sink, _lame_fake_sink, nullptr);
	}
}


bool GSTPlaybackPipeline::set_uri(gchar* uri) {

	if(!uri) return false;

	gst_element_set_state(_pipeline, GST_STATE_NULL);
	g_object_set(G_OBJECT(_audio_src), "uri", uri, nullptr);

	return true;
}


void GSTPlaybackPipeline::start_timer(qint64 play_ms) {

	if(play_ms > 0) {
		_timer->start(play_ms);
	}

	else {
		play();
	}
}


void GSTPlaybackPipeline::set_eq_band(QString band_name, double val) {

	g_object_set(G_OBJECT(_equalizer), band_name.toUtf8().data(), val, nullptr);
}


void GSTPlaybackPipeline::set_sr_path(QString path){

	if(!_file_sink) return;
	if(path == _sr_path) return;

	_sr_path = path;

	bool success = true;
	GstState state;
	bool nulled;

	gst_element_get_state(_pipeline, &state, nullptr, GST_SECOND);


	nulled=false;
	if(path.isEmpty()){

		if(!gst_bin_get_by_name(GST_BIN(_pipeline), "sr_fakesink")){
			gst_element_set_state(_pipeline, GST_STATE_NULL);
			gst_bin_add(GST_BIN(_pipeline), _file_fake_sink);
			nulled = true;
		}

		if(gst_bin_get_by_name(GST_BIN(_pipeline), "sr_filesink")){
			gst_element_set_state(_pipeline, GST_STATE_NULL);
			gst_object_ref(_file_sink);
			gst_object_ref(_file_resampler);
			gst_object_ref(_file_lame);
			gst_element_unlink_many(_file_queue, _file_resampler, _file_lame, _file_sink, nullptr);
			gst_bin_remove_many(GST_BIN(_pipeline), _file_resampler, _file_lame, _file_sink, nullptr);
			nulled = true;
		}

		GstPad* pad = gst_element_get_static_pad(_file_fake_sink, "sink");
		if(pad && !gst_pad_is_linked(pad)){
			success = gst_element_link(_file_queue, _file_fake_sink);
		}
	}

	else{

		gst_element_set_state(_pipeline, GST_STATE_NULL);
		nulled = true;

		if(gst_bin_get_by_name(GST_BIN(_pipeline), "sr_fakesink")){
			gst_object_ref(_file_fake_sink);
			gst_element_unlink(_file_queue, _file_fake_sink);
			gst_bin_remove(GST_BIN(_pipeline), _file_fake_sink);
		}

		if(!gst_bin_get_by_name(GST_BIN(_pipeline), "sr_filesink")){
			gst_bin_add_many(GST_BIN(_pipeline), _file_resampler, _file_lame, _file_sink, nullptr);
			success = gst_element_link_many(_file_queue, _file_resampler, _file_lame, _file_sink, nullptr);
		}

		g_object_set(G_OBJECT(_file_sink), "location", path.toUtf8().data(), nullptr);

	}

	if(nulled){
		gst_element_set_state(_pipeline, state);
	}

	Q_UNUSED(success);
}
