/* Playlist.cpp */

/* Copyright (C) 2011  Lucio Carreras
 *
 * This file is part of sayonara player
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Apr 6, 2011
 *      Author: luke
 */

#include "PlaylistHandler.h"
#include "StdPlaylist.h"
#include "StreamPlaylist.h"


#include "Helper/DirectoryReader/DirectoryReader.h"
#include "Helper/Parser/PlaylistParser.h"
#include "Components/TagEdit/MetaDataChangeNotifier.h"

PlaylistHandler::PlaylistHandler(QObject * parent) :
	QObject (parent),
	SayonaraClass()
{

	MetaDataChangeNotifier* md_change_notifier = MetaDataChangeNotifier::getInstance();

	_db = CDatabaseConnector::getInstance();
	_engine = GSTEngineHandler::getInstance();
	_play_manager = PlayManager::getInstance();
	_playlist_db_connector = PlaylistDBConnector::getInstance();

	_cur_playlist_idx = -1;
	_active_playlist_idx = -1;

	connect(_play_manager,	&PlayManager::sig_playstate_changed, this, &PlaylistHandler::playstate_changed);
	connect(_play_manager, &PlayManager::sig_next, this, &PlaylistHandler::next);
	connect(_play_manager, &PlayManager::sig_previous, this, &PlaylistHandler::previous);

	connect(_engine, &GSTEngineHandler::sig_md_changed, this, &PlaylistHandler::md_changed_single);
	connect(_engine, &GSTEngineHandler::sig_dur_changed, this, &PlaylistHandler::md_changed_single);

	connect(md_change_notifier, &MetaDataChangeNotifier::sig_metadata_changed, this, &PlaylistHandler::md_changed);
}

PlaylistHandler::~PlaylistHandler() {

}

void PlaylistHandler::emit_playlist_created(PlaylistPtr pl){

	if(!pl){
		pl = get_current();
	}

	if(pl->is_temporary()){

		if(pl->get_db_id() >= 0){
			pl->save();
		}

		else{
			pl->insert_temporary_into_db();
		}
	}

	emit sig_playlist_created( pl );
}


void PlaylistHandler::emit_cur_track_changed(PlaylistPtr pl){

	MetaData md;
	bool success;
	int cur_track_idx;
	int playlist_idx;
	int playlist_id;

	if(!pl){
		pl = get_active();
	}

	success = pl->get_cur_track(md);
	cur_track_idx = pl->get_cur_track_idx();
	playlist_idx = pl->get_idx();

	if(!success || cur_track_idx == -1){
		_play_manager->change_track_idx(-1);
		return;
	}

	playlist_id = pl->get_db_id();
	if(playlist_id == -1){
		pl->insert_temporary_into_db();
		playlist_id = pl->get_db_id();
	}

	_settings->set(Set::PL_LastPlaylist, playlist_id);

	_play_manager->change_track(md);
	_play_manager->change_track_idx(cur_track_idx);

	emit sig_cur_track_idx_changed( cur_track_idx,	playlist_idx );
}


int PlaylistHandler::load_old_playlists(){

	int last_playlist_idx=0;
	int last_track_idx=0;
	PlaylistPtr playlist;

	PlaylistLoader loader;

	_active_playlist_idx = -1;
	_cur_playlist_idx = -1;

	// returns -1 if fail, else the correct idx
	last_playlist_idx = loader.get_last_playlist_idx();
	last_track_idx = loader.get_last_tracks_idx();

	// no playlists found
	if( loader.get_playlist_count() == 0 ){
		create_empty_playlist(request_new_playlist_name());
		change_playlist_index(0);
		set_active_idx(0);
		_play_manager->stop();
		return 0;
	}

	// add playlists
	for(const CustomPlaylist& pl : loader.get_playlists()){
		create_playlist(pl);
	}

	playlist = get_playlist(last_playlist_idx, nullptr);

	// playlist not found -> track not found
	if(playlist == nullptr){
		last_track_idx = 0;
		last_playlist_idx = 0;
	}

	else{
		int n_tracks = playlist->get_size();

		// track not found
		if( !between(last_track_idx, 0, n_tracks) )
		{
			last_track_idx = 0;
		}
	}

	change_playlist_index(last_playlist_idx);
	set_active_idx(last_playlist_idx);
	change_track(last_track_idx, last_playlist_idx);

	if(!_settings->get(Set::PL_StartPlaying)){
		_play_manager->pause();
	}

	else{
		_play_manager->play();
	}

	return _playlists.size();
}


PlaylistPtr PlaylistHandler::new_playlist(Playlist::Type type, int playlist_idx, QString name) {

    switch(type) {

		case Playlist::Type::Stream:
			return PlaylistPtr(new StreamPlaylist(playlist_idx, name));

		case Playlist::Type::Std:
		default:
			return PlaylistPtr(new StdPlaylist(playlist_idx, name));
	}

	return nullptr;
}

// create a playlist, where metadata is already available
int PlaylistHandler::create_playlist(const MetaDataList& v_md, QString name, bool temporary, Playlist::Type type) {

	int idx;
	PlaylistPtr current;

	idx = add_new_playlist(name, temporary, type);
	change_playlist_index(idx);

	current = get_current();
	current->create_playlist(v_md);
	current->set_temporary( current->is_temporary() && temporary ) ;

	emit_playlist_created();
	return _cur_playlist_idx;
}


// create a new playlist, where only filepaths are given
// Load Folder, Load File...
int PlaylistHandler::create_playlist(const QStringList& pathlist, QString name, bool temporary, Playlist::Type type) {

	DirectoryReader reader;
	MetaDataList v_md = reader.get_md_from_filelist(pathlist);
	return create_playlist(v_md, name, temporary, type);
}


int PlaylistHandler::create_playlist(const QString& dir, QString name, bool temporary, Playlist::Type type) {

	QStringList lst;
	lst << dir;
	return create_playlist(lst, name, temporary, type);
}

int PlaylistHandler::create_empty_playlist(const QString name, bool temporary){

	MetaDataList v_md;
	return create_playlist(v_md, name, temporary);
}


int PlaylistHandler::create_playlist(const CustomPlaylist& cpl){

	int idx;
	PlaylistPtr pl;

	idx = create_playlist(cpl.tracks, cpl.name, cpl.is_temporary);
	pl = get_playlist(idx, nullptr);
	pl->set_db_id(cpl.id);

	return idx;
}


void PlaylistHandler::save_playlist(QString filename, bool relative) {

	PlaylistParser::save_playlist(filename, get_current()->get_playlist(), relative);
}


PlaylistDBInterface::SaveAsAnswer PlaylistHandler::save_cur_playlist(){
	return save_playlist(_cur_playlist_idx);
}

PlaylistDBInterface::SaveAsAnswer PlaylistHandler::save_playlist(int idx){

	sp_log(Log::Debug) << "Handler: save_playlist";

	PlaylistDBInterface::SaveAsAnswer ret;

	if( !between(idx, 0, _playlists.size()) ){
		return PlaylistDBInterface::SaveAsAnswer::Error;
	}

	ret = _playlists[idx]->save();

	emit sig_playlists_changed();

	return ret;
}


PlaylistDBInterface::SaveAsAnswer PlaylistHandler::save_cur_playlist_as(const QString& name, bool force_override){
	if(name.isEmpty()){
		return PlaylistDBInterface::SaveAsAnswer::Error;
	}

	return save_playlist_as(_cur_playlist_idx, name, force_override);
}

PlaylistDBInterface::SaveAsAnswer PlaylistHandler::save_playlist_as(int idx, const QString& name, bool force_override){

	int new_idx;
	PlaylistPtr pl, new_pl;
	PlaylistDBInterface::SaveAsAnswer ret;

	// no empty playlists
	if(name.isEmpty()){
		return PlaylistDBInterface::SaveAsAnswer::Error;
	}

	// get playlist we wanna save
	pl = get_playlist(idx, nullptr);
	if( pl == nullptr ){
		return PlaylistDBInterface::SaveAsAnswer::Error;
	}

	ret = pl->save_as(name, force_override);
	if(ret != PlaylistDBInterface::SaveAsAnswer::Success){
		return ret;
	}

	pl->set_temporary(false);

	new_idx = create_playlist(pl->get_playlist(), name, false);
	new_pl = get_playlist(new_idx, nullptr);
	new_pl->change_track(pl->get_cur_track_idx());

	close_playlist(idx);

	if(_cur_playlist_idx == idx){
		_cur_playlist_idx = new_pl->get_idx();
	}

	if(_active_playlist_idx == idx){
		_active_playlist_idx = new_pl->get_idx();
		int db_id = get_active()->get_db_id();
		int track_idx = get_active()->get_cur_track_idx();
		_settings->set(Set::PL_LastPlaylist, db_id);
		_settings->set(Set::PL_LastTrack, track_idx);
	}

	emit sig_playlists_changed();

	return PlaylistDBInterface::SaveAsAnswer::Success;
}

void PlaylistHandler::delete_playlist(int idx){

	bool success = _playlists[idx]->remove_from_db();

	if(success){
		emit sig_playlists_changed();
	}
}


void PlaylistHandler::md_changed_single(const MetaData& md) {

	for( PlaylistPtr& pl : _playlists ){
		IdxList idx;
		idx << pl->find_tracks(md.filepath());

		for(int i : idx){
			pl->replace_track(i, md);
		}
	}

	emit_playlist_created();
}


void PlaylistHandler::md_changed(const MetaDataList& v_md_old, const MetaDataList& v_md_new) {
	for(PlaylistPtr& pl : _playlists){
		pl->metadata_changed(v_md_old, v_md_new);
	}

	emit_playlist_created();
}


// GUI -->
void PlaylistHandler::clear_playlist() {

	PlaylistPtr current = get_current();
	current->clear();

	emit_playlist_created(current);
}

void PlaylistHandler::playstate_changed(PlayManager::PlayState state){

	switch(state){
		case PlayManager::PlayState::Playing:
			played();
			break;
		case PlayManager::PlayState::Paused:
			paused();
			break;
		case PlayManager::PlayState::Stopped:
			stopped();
			break;

		default:
			return;
	}
}

void PlaylistHandler::played(){

	get_active()->play();
}

void PlaylistHandler::paused(){

	get_active()->pause();
}

void PlaylistHandler::stopped(){

	get_active()->stop();
}


void PlaylistHandler::next() {
	get_active()->next();
	emit_cur_track_changed();
}


void PlaylistHandler::previous() {
	get_active()->bwd();
	emit_cur_track_changed();
}


void PlaylistHandler::set_active_idx(int idx){
	PlaylistPtr active;

	_active_playlist_idx = idx;
	active = get_active();


	if(active == nullptr){
		return;
	}

	_active_playlist_idx = active->get_idx();

	_settings->set(Set::PL_LastPlaylist, get_active()->get_db_id());
}


void PlaylistHandler::change_track(int idx, int playlist_idx) {

	if(_active_playlist_idx == -1){
		_active_playlist_idx = playlist_idx;
	}

	if(playlist_idx != _active_playlist_idx && playlist_idx >= 0){
		get_active()->stop();
		set_active_idx(playlist_idx);
	}

	get_active()->change_track(idx);

	emit_cur_track_changed();
}


void PlaylistHandler::selection_changed(const IdxList& lst) {

	PlaylistPtr current = get_current();

	current->selection_changed(lst);

	emit sig_selection_changed( current->get_selected_tracks() );

}

void PlaylistHandler::play_next(const MetaDataList& v_md) {

	PlaylistPtr active = get_active();

	active->insert_tracks(v_md, active->get_cur_track_idx() + 1);
	emit_playlist_created(active);
}


void PlaylistHandler::insert_tracks(const MetaDataList& v_md, int row, int pl_idx) {

	PlaylistPtr pl = get_playlist(pl_idx, get_current());

	pl->insert_tracks(v_md, row);
	emit_playlist_created(pl);
}


void PlaylistHandler::append_tracks(const MetaDataList& v_md, int pl_idx) {

	PlaylistPtr pl = get_playlist(pl_idx, get_current());

	pl->append_tracks(v_md);
	emit_playlist_created(pl);
}

void PlaylistHandler::remove_rows(const IdxList& rows, int pl_idx) {

	PlaylistPtr pl = get_playlist(pl_idx, get_current());

	pl->delete_tracks(rows);
	emit_playlist_created(pl);
}

void PlaylistHandler::move_rows(const IdxList& rows, int tgt_idx, int pl_idx) {

	PlaylistPtr pl = get_playlist(pl_idx, get_current());

	pl->move_tracks(rows, tgt_idx);
	emit_playlist_created(pl);
}


QString PlaylistHandler::request_new_playlist_name() const
{
	return PlaylistDBInterface::request_new_db_name();
}


int PlaylistHandler::add_new_playlist(const QString& name, bool temporary, Playlist::Type type){

	int idx = exists(name);
	PlaylistPtr pl;

	if(idx >= 0) {
		return idx;
	}

	idx = _playlists.size();

	pl = new_playlist(type, _playlists.size(), name);
	pl->set_temporary(temporary);

	_playlists.append(pl);

	emit sig_new_playlist_added(idx, name);

	return idx;
}


int PlaylistHandler::change_playlist_index(int idx){

	PlaylistPtr current;

	_cur_playlist_idx = idx;

	current = get_current();

	if(_play_manager->get_play_state() == PlayManager::PlayState::Stopped){
		set_active_idx(idx);
	}

	emit_playlist_created();

	emit sig_playlist_mode_changed( current->get_playlist_mode());
	emit sig_playlist_index_changed(idx, current->is_temporary());

	return _cur_playlist_idx;
}

void PlaylistHandler::close_playlist(int idx){

	bool was_active = (idx == _active_playlist_idx);
	bool was_current = (idx == _cur_playlist_idx);

	if(_playlists[idx]->is_temporary()){
		QString name = _playlists[idx]->get_name();
		_playlist_db_connector->delete_playlist( name );
	}

	_playlists.removeAt(idx);

	if( was_current ){
		_cur_playlist_idx = 0;
	}

	else if(_cur_playlist_idx > idx){
		_cur_playlist_idx --;
	}

	if( was_active ){
		set_active_idx(0);
	}

	else if(_active_playlist_idx > idx){
		_active_playlist_idx --;
	}

	for(PlaylistPtr pl : _playlists){
		if(pl->get_idx() >= idx){
			pl->set_idx(pl->get_idx() - 1);
		}
	}

	if(was_active){
		_settings->set(Set::PL_LastPlaylist, -1);
		_settings->set(Set::PL_LastTrack, -1);
	}
	else{
		_settings->set(Set::PL_LastPlaylist, get_active()->get_db_id());
	}

	emit_playlist_created();

	emit sig_playlist_closed(idx);
}

PlaylistPtr PlaylistHandler::get_playlist(int idx, PlaylistPtr fallback) const
{
	if(! between(idx, 0, _playlists.size())){
		return fallback;
	}

	return _playlists[idx];
}

PlaylistPtr PlaylistHandler::get_active()
{
	// assure we have at least one playlist
	if(_playlists.size() == 0){
		_active_playlist_idx = add_new_playlist(request_new_playlist_name(), true, Playlist::Type::Std);
		_cur_playlist_idx = _active_playlist_idx;
	}
	// assure valid idx
	if( !between(_active_playlist_idx, 0,_playlists.size()) ){
		_active_playlist_idx = 0;
	}

	return get_playlist(_active_playlist_idx, nullptr);
}

PlaylistPtr PlaylistHandler::get_current()
{

	// assure we have at least one playlist
	if(_playlists.size() == 0){
		_cur_playlist_idx = add_new_playlist(request_new_playlist_name(), true, Playlist::Type::Std);
		_active_playlist_idx = _cur_playlist_idx;
	}
	// assure valid idx
	if( !between(_cur_playlist_idx, 0, _playlists.size())){
		_cur_playlist_idx = 0;
	}

	return get_playlist(_cur_playlist_idx, nullptr);
}

int	PlaylistHandler::get_active_idx() const
{
	return _active_playlist_idx;
}

int	PlaylistHandler::get_current_idx() const
{
	return _cur_playlist_idx;
}

int PlaylistHandler::exists(const QString& name) const
{
	if(name.isEmpty()) {
		return _cur_playlist_idx;
	}

	for(const PlaylistPtr& pl : _playlists){

		if(pl->get_name().compare(name, Qt::CaseInsensitive) == 0){
			return pl->get_idx();
		}
	}

	return -1;
}

MetaDataList PlaylistHandler::get_tracks(PlaylistIndex which){

	if(which == PlaylistIndex::Current){
		return get_current()->get_playlist();
	}

	else{
		return get_active()->get_playlist();
	}
}

//
void PlaylistHandler::similar_artists_available(const IDList& artists) {

	if(artists.isEmpty()){
		return;
	}

	Playlist::Type type;
	PlaylistMode plm;
	RandomGenerator rnd;
	MetaData md;
	int cur_artist_idx;
	bool is_track_already_in;
	IDList artists_copy;

	type = get_active()->get_type();
	plm = _settings->get(Set::PL_Mode);
	if( (type != Playlist::Type::Std) || !plm.dynamic ) {
		return;
	}

	artists_copy = Helper::randomize_list(artists);

	cur_artist_idx = 0;
	is_track_already_in = false;

	do {
		int artist_id;
		MetaDataList vec_tracks;

		artist_id = artists_copy[cur_artist_idx];
		_db->getAllTracksByArtist(artist_id, vec_tracks);

		// give each artist several trys
		int max_rounds = vec_tracks.size();
		for(int rounds=0; rounds < max_rounds; rounds++) {

			int rnd_track = rnd.get_number(0, vec_tracks.size()- 1);

			md = vec_tracks[rnd_track];

			// search playlist
			if( ! get_active()->find_tracks(md.id).isEmpty() ) {
				is_track_already_in = true;
			}

			else break;
		}

		cur_artist_idx++;

	} while(is_track_already_in && cur_artist_idx < artists_copy.size());


	if(md.id < 0) {
		return;
	}

	if(!is_track_already_in){
		get_active()->append_track(md);
		emit_playlist_created(get_active());
	}
}
