/* CDatabaseTracks.cpp */

/* Copyright (C) 2011  Lucio Carreras
 *
 * This file is part of sayonara player
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "DatabaseAccess/CDatabaseConnector.h"
#include "Helper/Helper.h"

#include <QSqlResult>
#include <QElapsedTimer>


using namespace Sort;

#define TRACK_SELECTOR QString("SELECT ") + \
	"tracks.trackID AS trackID, " \
	"tracks.title AS trackTitle, "	\
	"tracks.length AS trackLength, " \
	"tracks.year AS trackYear, " \
	"tracks.bitrate AS trackBitrate, " \
	"tracks.filename AS trackFilename, " \
	"tracks.track AS trackNum, " \
	"albums.albumID AS albumID, " \
	"artists.artistID AS artistID, " \
	"albums.name AS albumName, " \
    "artists.name AS artistName, " \
    "tracks.genre AS genrename, "\
    "tracks.filesize AS filesize, " \
    "tracks.discnumber AS discnumber, " \
    "tracks.rating AS rating " \
    "FROM tracks " \
    "INNER JOIN albums ON tracks.albumID = albums.albumID " \
	"INNER JOIN artists ON tracks.artistID = artists.artistID " \
	"WHERE filetype is null "


bool CDatabaseConnector::db_fetch_tracks(SayonaraQuery& q, MetaDataList& result) {
	return _db_tracks->db_fetch_tracks(q, result);
}


bool CDatabaseConnector::getMultipleTracksByPath(QStringList& paths, MetaDataList& v_md) {
	return _db_tracks->getMultipleTracksByPath(paths, v_md);
}


MetaData CDatabaseConnector::getTrackByPath(const QString& path) {
	return _db_tracks->getTrackByPath(path);
}

MetaData CDatabaseConnector::getTrackById(int id) {
	return _db_tracks->getTrackById(id);
}


bool CDatabaseConnector::getTracksFromDatabase (MetaDataList& v_md, SortOrder sort) {
	return _db_tracks->getTracksFromDatabase(v_md, sort);
}

bool CDatabaseConnector::getAllTracksByAlbum(int album, MetaDataList& v_md, Filter filter, SortOrder sort, int discnumber) {
	return _db_tracks->getAllTracksByAlbum(album, v_md, filter, sort, discnumber);
}

bool CDatabaseConnector::getAllTracksByAlbum(QList<int> albums, MetaDataList& v_md, Filter filter, SortOrder sort) {
	return _db_tracks->getAllTracksByAlbum(album, v_md, filter, sort);
}

bool CDatabaseConnector::getAllTracksByArtist(int artist, MetaDataList& v_md, Filter filter, SortOrder sort) {
	return _db_tracks->getAllTracksByArtist(artist, v_md, filter, sort);
}

bool CDatabaseConnector::getAllTracksByArtist(QList<int> artists, MetaDataList& v_md, Filter filter, SortOrder sort) {
	return _db_tracks->getAllTracksByArtist(artists, v_md, filter, sort);
}


bool CDatabaseConnector::getAllTracksBySearchString(Filter filter, MetaDataList& result, SortOrder sort) {
	return _db_tracks->getAllTracksBySearchString(filter, result, sort);
}

bool CDatabaseConnector::deleteTrack(int id){
	return _db_tracks->deleteTrack(id);
}


bool CDatabaseConnector::deleteTracks(const QList<int>& ids){
	return _db_tracks->deleteTrack(ids);
}


bool CDatabaseConnector::deleteTracks(const MetaDataList& v_md) {

	int success = 0;

	_database.transaction();

		for(const MetaData& md : v_md){
			if( deleteTrack(md.id) ){
				success++;
			};
		}

	_database.commit();

	sp_log(LogInfo) << "Deleted " << success << " of " << v_md.size() << " tracks";

	return (success == v_md.size());
}

bool CDatabaseConnector::deleteInvalidTracks(){

	bool success;

	MetaDataList v_md;
	QMap<QString, int> map;
	QList<int> to_delete;
	MetaDataList v_md_update;

	DB_RETURN_NOT_OPEN_BOOL(_database);

	SayonaraQuery q(_database);


	if(!getTracksFromDatabase(v_md)){
		sp_log(LogError) << "Cannot get tracks from db";
		return false;
	}

	int idx = 0;
	for(const MetaData& md : v_md){

		if(map.contains(md.filepath())){
			sp_log(LogWarning) << "found double path: " << md.filepath();
			int old_idx = map[md.filepath()];
			to_delete << md.id;
			v_md_update << v_md[old_idx];
		}

		else{
			map.insert(md.filepath(), idx);
		}

		idx++;
	}

	sp_log(LogDebug) << "Will delete " << to_delete;
	success = deleteTracks(to_delete);
	sp_log(LogDebug) << "delete tracks " << success;

	success = deleteTracks(v_md_update);
	sp_log(LogDebug) << "delete other tracks " << success;

	success = storeMetadata(v_md_update);
	sp_log(LogDebug) << "update tracks " << success;

	return false;
}


bool CDatabaseConnector::updateTrack(const MetaData& md) {

	if(md.id == -1) return false;

	DB_RETURN_NOT_OPEN_BOOL(_database);

	SayonaraQuery q (_database);

	q.prepare("UPDATE Tracks "
			  "SET albumID=:albumID, "
				   "artistID=:artistID, "
				   "title=:title, "
				   "year=:year, "
				   "length=:length, "
				   "bitrate=:bitrate, "
				   "track=:track, "
				   "genre=:genre, "
				   "filesize=:filesize, "
				   "discnumber=:discnumber, "
				   "cissearch=:cissearch, "
				   "rating=:rating "
				"WHERE TrackID = :trackID;");

	q.bindValue(":albumID",QVariant(md.album_id));
	q.bindValue(":artistID",QVariant(md.artist_id));
	q.bindValue(":title",QVariant(md.title));
	q.bindValue(":track",QVariant(md.track_num));
	q.bindValue(":length", QVariant(md.length_ms));
	q.bindValue(":bitrate", QVariant(md.bitrate));
	q.bindValue(":year",QVariant(md.year));
	q.bindValue(":trackID", QVariant(md.id));
	q.bindValue(":genre", QVariant(md.genres.join(",")));
	q.bindValue(":filesize", QVariant(md.filesize));
	q.bindValue(":discnumber", QVariant(md.discnumber));
	q.bindValue(":rating", QVariant(md.rating));
	q.bindValue(":cissearch", QVariant(md.title.toLower()));

	if (!q.exec()) {
		q.show_error(QString("Cannot update track ") + md.filepath());
		return false;
	}

	return true;
}

bool CDatabaseConnector::updateTracks(const MetaDataList& lst) {

	bool success;
	int n_files = 0;

    _database.transaction();
	for(const MetaData& md : lst){

		if(updateTrack(md)){
			n_files++;
		}
	}

	success = _database.commit();

	return success && (n_files == lst.size());
}

int CDatabaseConnector::insertTrackIntoDatabase (const MetaData & md, int artistID, int albumID) {

	DB_RETURN_NOT_OPEN_INT(_database);

	SayonaraQuery q (_database);

	MetaData md_tmp = getTrackByPath( md.filepath() );

	if( md_tmp.id >= 0 ) {
		MetaData track_copy = md;
		track_copy.id = md_tmp.id;
		track_copy.artist_id = artistID;
		track_copy.album_id = albumID;

		if( updateTrack(track_copy) ){
			return md_tmp.id;
		}
	}

	sp_log(LogInfo) << "insert new track: " << md.filepath();

	QString querytext = QString("INSERT INTO tracks ") +
                "(filename,albumID,artistID,title,year,length,track,bitrate,genre,filesize,discnumber,cissearch) " +
				"VALUES "+
                "(:filename,:albumID,:artistID,:title,:year,:length,:track,:bitrate,:genre,:filesize,:discnumber,:cissearch); ";

	q.prepare(querytext);

	q.bindValue(":filename", md.filepath());
	q.bindValue(":albumID", albumID);
	q.bindValue(":artistID",artistID);
	q.bindValue(":length", md.length_ms);
	q.bindValue(":year", md.year);
	q.bindValue(":title", md.title);
	q.bindValue(":track", md.track_num);
	q.bindValue(":bitrate", md.bitrate);
	q.bindValue(":genre", md.genres.join(","));
	q.bindValue(":filesize", md.filesize);
	q.bindValue(":discnumber", md.discnumber);
	q.bindValue(":cissearch", md.title.toLower());

    if (!q.exec()) {

		q.show_error(QString("Cannot insert track into database ") + md.filepath());
		return -1;
    }

	return getTrackByPath(md.filepath()).id;
}



