/* GUI_Player.cpp */

/* Copyright (C) 2011-2015  Lucio Carreras
 *
 * This file is part of sayonara player
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "GUI_Player.h"
#include "GUI_TrayIcon.h"
#include "GUI/AlternativeCovers/GUI_AlternativeCovers.h"

#include "Helper/globals.h"
#include "Helper/Style/Style.h"
#include "Helper/WebAccess/AsyncWebAccess.h"

#include "Components/CoverLookup/CoverLookup.h"
#include "Components/StreamPlugins/LastFM/LastFM.h"

#include <QFileDialog>
#include <QPalette>

GUI_Player* obj_ref = nullptr;

#ifdef Q_OS_UNIX
#include <csignal>

void signal_handler(int sig) {

	if(sig == SIGWINCH && obj_ref) {
		sp_log(Log::Debug) << "show everything";
		obj_ref->setHidden(false);
		obj_ref->showNormal();
		obj_ref->activateWindow();
	}

	else if(sig == SIGTERM && obj_ref) {
		sp_log(Log::Debug) << "extern close event";
		obj_ref->really_close();
	}
	else {

		sp_log(Log::Debug) << "signal " << sig << " received";
	}
}

#endif

GUI_Player::GUI_Player(QTranslator* translator, QWidget *parent) :
	SayonaraMainWindow(parent),
	GlobalMessageReceiver("Player Main Window"),
	Ui::Sayonara()
{
	setupUi(this);

	GlobalMessage::getInstance()->register_receiver(this);

	init_gui();

	_translator = translator;
	_engine = GSTEngineHandler::getInstance();
	_play_manager = PlayManager::getInstance();

	set_cur_pos_ms(_play_manager->get_init_position_ms());

	QString version = _settings->get(Set::Player_Version);

	_awa_version = new AsyncWebAccess(this);
	_awa_translators = new AsyncWebAccess(this);

	lab_sayonara->setText(tr("Sayonara Player"));
	lab_version->setText( version );
	lab_writtenby->setText(tr("Written by") + " Lucio Carreras");
	lab_copyright->setText(tr("Copyright") + " 2011-2015");

	_md_available = false;
	_mute = false;

	ui_playlist = nullptr;
	_local_library = nullptr;
	sli_buffer->setVisible(false);

	ui_streamrecorder = new GUI_StreamRecorder(this);
	ui_broadcasting = new GUI_BroadcastSetup(this);
	ui_notifications = new GUI_Notifications(this);
	ui_StartupDialog = new GUI_StartupDialog(this);
	ui_language_chooser= new GUI_LanguageChooser(this);
	ui_lastfm = new GUI_LastFM(this);

	_skin_suffix = "";

	_cov_lookup = new CoverLookup(this);
	_ui_alternative_covers = new GUI_AlternativeCovers(this->centralWidget());

	action_min2tray->setChecked(_settings->get(Set::Player_Min2Tray));

	bool one_instance = _settings->get(Set::Player_OneInstance);
	action_only_one_instance->setChecked(one_instance);

	bool show_small_items = _settings->get(Set::PL_SmallItems);
	action_smallPlaylistItems->setChecked(show_small_items);

	bool show_only_tracks = _settings->get(Set::Lib_OnlyTracks);
	action_showOnlyTracks->setChecked(show_only_tracks);

	QSizePolicy p = library_widget->sizePolicy();
	_library_stretch_factor = p.horizontalStretch();

	bool show_library = _settings->get(Set::Lib_Show);
	action_viewLibrary->setChecked(show_library);
	this->show_library(show_library);

	bool live_search = _settings->get(Set::Lib_LiveSearch);
	action_livesearch->setChecked(live_search);

	bool notify_new_version = _settings->get(Set::Player_NotifyNewVersion);
	action_notifyNewVersion->setChecked(notify_new_version);

	bool is_fullscreen = _settings->get(Set::Player_Fullscreen);
	bool is_maximized = _settings->get(Set::Player_Maximized);

	int volume = _settings->get(Set::Engine_Vol);
	volume_changed(volume);

	action_Dark->setChecked(_settings->get(Set::Player_Style) == 1);
	skin_changed();

	if(!is_fullscreen & !is_maximized) {

		QSize size = _settings->get(Set::Player_Size);
		QPoint pos = _settings->get(Set::Player_Pos);
		QRect rect = this->geometry();

		rect.setX(pos.x());
		rect.setY(pos.y());
		rect.setWidth(size.width());
		rect.setHeight(size.height());
		this->setGeometry(rect);
	}

	_ui_library_width = 600;

	ui_libpath = new GUI_LibraryPath( library_widget );

	/* tray actions */
	setup_tray_actions();

	// signals and slots
	setup_connections();


	if(_settings->get(Set::Player_NotifyNewVersion)){
		_awa_version->run("http://sayonara-player.com/current_version");
	}

	_awa_translators->run("http://sayonara-player.com/translators");

	plugin_widget->resize(plugin_widget->width(), 0);

	stopped();

	REGISTER_LISTENER(Set::Lib_Path, _sl_libpath_changed);
	REGISTER_LISTENER(Set::Engine_SR_Active, _sl_sr_active_changed);
	REGISTER_LISTENER_NO_CALL(Set::Player_Fullscreen, _sl_fullscreen_toggled);
	REGISTER_LISTENER_NO_CALL(SetNoDB::Player_Quit, really_close);
}


GUI_Player::~GUI_Player() {
	sp_log(Log::Info) << "closing player...";
}


void GUI_Player::language_changed() {

	QString language = _settings->get(Set::Player_Language);
	_translator->load(language, Helper::getSharePath() + "translations/");

	retranslateUi(this);

	QList<PlayerPlugin*> all_plugins = _pph->get_all_plugins();
	QList<QAction*> actions = menuView->actions();

	for(QAction* action : actions) {
		if(!action->data().isNull()) {
			menuView->removeAction(action);
		}
	}

	actions.clear();

	for(PlayerPlugin* p : all_plugins) {
		QAction* action = p->getAction();
		action->setData(p->getName());
		actions << action;
	}

	menuView->insertActions(action_Dark, actions);
	menuView->insertSeparator(action_Dark);
}


QAction* GUI_Player::create_actions(QList<QKeySequence>& seq_list) {
	QAction* action = new QAction(this);

	action->setShortcuts(seq_list);
	action->setShortcutContext(Qt::ApplicationShortcut);
	this->addAction(action);

	return action;
}

QAction* GUI_Player::create_action(QKeySequence seq) {
	QList<QKeySequence> seq_list;
	seq_list << seq;
	return create_actions(seq_list);
}


void GUI_Player::init_gui() {

	btn_mute->setIcon(Helper::getIcon("vol_1"));

	action_viewLibrary->setText(tr("&Library"));
	btn_rec->setVisible(false);

	action_Fullscreen->setShortcut(QKeySequence("F11"));
	action_Dark->setShortcut(QKeySequence("F10"));
}


// new track
void GUI_Player::track_changed(const MetaData & md) {

	_md = md;
	_md_available = true;

	lab_sayonara->hide();
	lab_title->show();

	lab_version->hide();
	lab_artist->show();

	lab_writtenby->hide();
	lab_album->show();

	lab_copyright->hide();
	lab_rating->show();

	set_album_label();

	lab_artist->setText(Helper::elide_text(_md.artist, lab_artist, 1));
	lab_title->setText(Helper::elide_text(_md.title, lab_title, 2));

	set_total_time_label(_md.length_ms);

	QString rating_text;


	if(_md.bitrate / 1000 > 0){
		rating_text = QString::number(_md.bitrate / 1000) + " kBit/s";
	}

	if(_md.filesize > 0){
		if(_md.bitrate / 1000 > 0){
			rating_text += ", ";
		}
		rating_text += QString::number( (double) (_md.filesize / 1024) / 1024.0, 'f', 2) + " MB";
	}

	lab_rating->setText(rating_text);
	lab_rating->setToolTip(rating_text);

	set_radio_mode( _md.radio_mode() );

	this->setWindowTitle(QString("Sayonara - ") + md.title);
	this->repaint();

	fetch_cover();
}

void GUI_Player::set_album_label(){

	QString str_year = QString::number(_md.year);
	QString album_name = _md.album;

	if(_md.year > 1000 && !_md.album.contains(str_year)){
		album_name += " (" + str_year + ")";
	}

	lab_album->setText(Helper::elide_text(album_name,lab_album, 1));
}

void GUI_Player::dur_changed(const MetaData& md) {

	if(_md != md){
		return;
	}

	set_total_time_label(md.length_ms);
}

void GUI_Player::br_changed(const MetaData& md) {

	QString rating_text;

	if(_md != md){
		return;
	}

	if(md.bitrate / 1000 > 0){
		rating_text = QString::number(md.bitrate / 1000) + " kBit/s";
	}

	if(md.filesize > 0){
		if(md.bitrate / 1000 > 0){
			rating_text += ", ";
		}
		rating_text += QString::number( (double) (md.filesize / 1024) / 1024.0, 'f', 2) + " MB";
	}

	lab_rating->setText(rating_text);
	lab_rating->setToolTip(rating_text);

}

void GUI_Player::md_changed(const MetaData& md) {

	if(_md != md && md.title == _md.title){
		return;
	}

	sp_log(Log::Debug) << "Metadata changed ";
	md.print();

	_md = md;
	_md.title = md.title;

	lab_title->setText( md.title );
}


// public slot:
// id3 tags have changed
void GUI_Player::id3_tags_changed(const MetaDataList& v_md_old, const MetaDataList& v_md_new) {

	IdxList idxs = v_md_old.findTracks(_md.filepath());
	if(idxs.isEmpty()) return;

	int idx = idxs[0];

	_md = v_md_new[idx];

	set_album_label();
	lab_artist->setText(_md.artist);
	lab_title->setText(_md.title);

	setWindowTitle(QString("Sayonara - ") + _md.title);
	fetch_cover();
}


void GUI_Player::skin_toggled(bool on){
	_settings->set(Set::Player_Style, (on ? 1 : 0) );
}

void GUI_Player::skin_changed() {

	bool dark = (_settings->get(Set::Player_Style) == 1);
	QString stylesheet = Style::get_style(dark);

	this->setStyleSheet(stylesheet);

	if (dark) 	_skin_suffix = QString("_dark");
	else 		_skin_suffix = QString("");

	setup_volume_button(volumeSlider->value());
}



/** TRAY ICON **/
void GUI_Player::setup_tray_actions() {

	_tray_icon = new GUI_TrayIcon(this);

	connect(_tray_icon, &GUI_TrayIcon::sig_stop_clicked, this, &GUI_Player::stop_clicked);
	connect(_tray_icon, &GUI_TrayIcon::sig_bwd_clicked, this, &GUI_Player::prev_clicked);
	connect(_tray_icon, &GUI_TrayIcon::sig_fwd_clicked, this, &GUI_Player::next_clicked);
	connect(_tray_icon, &GUI_TrayIcon::sig_mute_clicked, this, &GUI_Player::mute_button_clicked);
	connect(_tray_icon, &GUI_TrayIcon::sig_close_clicked, this, &GUI_Player::really_close);
	connect(_tray_icon, &GUI_TrayIcon::sig_play_clicked, this, &GUI_Player::play_clicked);
	connect(_tray_icon, &GUI_TrayIcon::sig_pause_clicked, this, &GUI_Player::play_clicked);
	connect(_tray_icon, &GUI_TrayIcon::sig_show_clicked, this, &GUI_Player::show_again);
	connect(_tray_icon, &GUI_TrayIcon::activated, this, &GUI_Player::tray_icon_activated);
	connect(_tray_icon, &GUI_TrayIcon::onVolumeChangedByWheel, this, &GUI_Player::change_volume_by_tick);

	_tray_icon->setPlaying(false);

	_tray_icon->show();
}


void GUI_Player::tray_icon_activated (QSystemTrayIcon::ActivationReason reason) {

	bool min_to_tray = _settings->get(Set::Player_Min2Tray);
	switch (reason) {

	case QSystemTrayIcon::Trigger:

		if(this->isHidden())
		{
			this->setHidden(false);
			this->activateWindow();
		}

		else if( this->isMinimized() ||
				 !this->isVisible() ||
				 !this->isActiveWindow())

		{
			QSize size = _settings->get(Set::Player_Size);
			this->setHidden(false);
			this->showMaximized();
			this->showNormal();
			this->activateWindow();
			this->resize(size);
		}

		else{
			if(min_to_tray){
				this->setHidden(true);
			}

			else{
				this->showMinimized();
			}
		}


		break;

	case QSystemTrayIcon::MiddleClick:
		break;

	default:
		break;
	}
}


void GUI_Player::set_playlist(GUI_Playlist* playlist) {

	if(!playlist) return;

	playlist->setParent(this->playlist_widget);
	ui_playlist = playlist;
}


void GUI_Player::set_libraries(LibraryPluginLoader* plugin_loader, int idx) {

	ui_libraries.clear();

	for(const LibraryContainer* container : plugin_loader->get_libraries()){

		if(!container) {
			continue;
		}

		GUI_AbstractLibrary* ui_library = container->get_ui();
		AbstractLibrary* library = container->get_logic();

		if(container->is_local()){
			_local_library = library;
		}

		connect(ui_library, &GUI_AbstractLibrary::sig_lib_idx_changed, this, &GUI_Player::set_cur_library);

		ui_library->setParent(library_widget);
		ui_library->hide();

		ui_libraries << ui_library;
	}

	_cur_library = ui_libraries[idx];

	if(_cur_library == nullptr) return;

	_cur_library->show();
	_cur_library->resize(library_widget->size());
}


void GUI_Player::set_cur_library(int idx){

	int lib_idx=0;
	for(GUI_AbstractLibrary* library : ui_libraries){
		if(library == nullptr) continue;

		disconnect(library, &GUI_AbstractLibrary::sig_lib_idx_changed, this, &GUI_Player::set_cur_library);
		library->set_index(lib_idx++);
		connect(library, &GUI_AbstractLibrary::sig_lib_idx_changed, this, &GUI_Player::set_cur_library);

		library->hide();
	}

	_cur_library = ui_libraries[idx];

	if(_cur_library == nullptr) return;

	_cur_library->show();
	_cur_library->resize(library_widget->size());
}


void GUI_Player::set_player_plugin_handler(PlayerPluginHandler* pph) {

	_pph = pph;

	QList<PlayerPlugin*> lst = _pph->get_all_plugins();
	QList<QAction*> actions;

	for(PlayerPlugin* p : lst) {
		QAction* action = p->getAction();
		action->setData(p->getName());
		// action is connected in Plugin itself
		p->setParent(plugin_widget);
		actions << action;
	}

	menuView->insertActions(action_Dark, actions);
	menuView->insertSeparator(action_Dark);

	connect(_pph, &PlayerPluginHandler::sig_show_plugin, this, &GUI_Player::show_plugin);
	connect(_pph, &PlayerPluginHandler::sig_hide_all_plugins, this, &GUI_Player::hide_all_plugins);
}

void GUI_Player::reload_library_allowed(bool b) {
	action_reloadLibrary->setEnabled(b);
}


/** LIBRARY AND PLAYLIST END **/


// prvt fct
void GUI_Player::set_radio_mode(RadioMode radio) {

	bool stream_ripper = _settings->get(Set::Engine_SR_Active);

	if(stream_ripper) {

		bool btn_rec_visible = (radio != RadioMode::Off);

		if(btn_rec_visible) {
			btn_play->setVisible(radio == RadioMode::Off);
			btn_rec->setVisible(radio != RadioMode::Off);
		}

		else {
			btn_rec->setVisible(radio != RadioMode::Off);
			btn_play->setVisible(radio == RadioMode::Off);
		}

		btn_play->setEnabled(radio == RadioMode::Off);
	}

	else{

		btn_rec->setVisible(false);
		btn_play->setVisible(true);
	}

	_tray_icon->set_enable_fwd(true);
	sli_progress->setEnabled( (_md.length_ms / 1000) > 0 );

	_play_manager->record(btn_rec->isChecked() && btn_rec->isVisible());

}


// public slot
void GUI_Player::_sl_sr_active_changed() {

	bool active = _settings->get(Set::Engine_SR_Active);

	if(active) {
		btn_play->setVisible(_md.radio_mode() == RadioMode::Off);
		btn_rec->setVisible(_md.radio_mode() != RadioMode::Off);
	}

	else{
		btn_play->setVisible(true);
		btn_play->setEnabled(_md.radio_mode() == RadioMode::Off);
		btn_rec->setVisible(false);
	}

	btn_rec->setChecked(false);
}


void GUI_Player::ui_loaded() {

#ifdef Q_OS_UNIX
	obj_ref = this;
	signal(SIGWINCH, signal_handler);
#endif

	QString lib_path = _settings->get(Set::Lib_Path);
	if(_cur_library){
		if(QFile::exists(lib_path)){
			ui_libpath->hide();
			_cur_library->show();
		}

		else{
			_cur_library->hide();
			ui_libpath->show();
		}
	}

	bool fullscreen = _settings->get(Set::Player_Fullscreen);
	bool maximized = _settings->get(Set::Player_Maximized);
	QPoint pos = _settings->get(Set::Player_Pos);
	QSize sz = _settings->get(Set::Player_Size);

	action_Fullscreen->setChecked(fullscreen);

	this->setGeometry(pos.x(), pos.y(), sz.width(), sz.height());

	if(fullscreen){
		this->showFullScreen();
	}

	else if(maximized){
		this->showMaximized();
	}

	else{
		this->showNormal();
	}


	QString shown_plugin = _settings->get(Set::Player_ShownPlugin);
	PlayerPlugin* p  = _pph->find_plugin(shown_plugin);
	show_plugin(p);

}



void GUI_Player::awa_version_finished(bool success) {

	if(!success) return;

	QString new_version(_awa_version->get_data());
	QString cur_version = _settings->get(Set::Player_Version);
	bool notify_new_version = _settings->get(Set::Player_NotifyNewVersion);
	bool dark = (_settings->get(Set::Player_Style) == 1);

	new_version = new_version.trimmed();

	sp_log(Log::Info) << "Newest Version: " << new_version;
	sp_log(Log::Info) << "This Version:   " << cur_version;

	QString link;
	LINK("http://sayonara-player.com", "http://sayonara-player.com", dark, link);

	if(new_version > cur_version && notify_new_version) {
		Message::info(tr("A new version is available!") + "<br />" +  link);
	}
}

void GUI_Player::awa_translators_finished(bool success) {
	if(!success) return;

	QString data(_awa_translators->get_data());
	QStringList translators = data.split('\n');

	_translators.clear();

	for(QString str : translators) {
		if(str.trimmed().size() > 0) {
			_translators.push_back(str);
		}
	}
}

void GUI_Player::notify_new_version_toggled(bool b) {
	_settings->set(Set::Player_NotifyNewVersion, b);
}

void GUI_Player::show_again(){
	if(this->isHidden())
	{
		this->setHidden(false);
		this->showNormal();
		this->activateWindow();
	}

	else if( this->isMinimized() ||
			 this->isHidden() ||
			 !this->isVisible() ||
			 !this->isActiveWindow())

	{
		QSize size = _settings->get(Set::Player_Size);
		this->setHidden(false);
		this->showMaximized();
		this->showNormal();
		this->activateWindow();
		this->resize(size);
	}
}

void GUI_Player::really_close() {

	this->hide();
	this->close();
}
